#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
extern "C" int abs(...);
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"
#include "detect_comm.h"
#include "indep.h"

//indep.cc
extern int find_and_replace(SgExpression *exp,SgVarRefExp *old,SgExpression * new_exp=NULL);
extern Distribute_def * get_distdef(SgSymbol * asym);
extern Alignment_def * get_aligndef(SgSymbol * asym);

extern int do_level;
extern SgFile *global_file;

extern void delete_dummy_array(SgStatement *);
extern void add_dummy_array(SgStatement *);

void change_indepdo_prochedr(SgStatement * s)
{
  SgProcHedrStmt *  prochedr = (SgProcHedrStmt *) s;
  int org_numberOfParameters = prochedr->numberOfParameters();

//  SgSymbol * parameters[10];
  SgSymbol **parameters = new SgSymbol *[org_numberOfParameters];

  int i;

  for (i=0; i<org_numberOfParameters; i++) 
    parameters[i]=prochedr->parameter(i);

  for (i=0; i<org_numberOfParameters; i++) 
  {
    SgSymbol * parameter=parameters[i];
	// scalar variable skip		
    if (isSgArrayType(parameter->type()) == NULL ) continue;
    SgSymbol * argstride = s->getSymbolInCurrentScope(mixstring, "arg_stride", parameter->identifier());
    prochedr->insertParameterAfter(*parameter, *argstride);
  }
  delete [] parameters;
}

SgExpression *ddot_sub_triplet(SgArrayRefExp *array,int i,int flag){
	SgSymbol *arraysym=array->symbol();
	SgExpression* sub=array->subscript(i);
	ASSERT(sub->variant()==DDOT);
	switch(flag){
		case 0:
			if(sub->lhs()->lhs()){
                                        return sub->lhs()->lhs();
                                }
                                else{
                                        return get_lbound(arraysym,i);
                                }

			break;
		case 1:
                                if(sub->lhs()->rhs()){
                                        return sub->lhs()->rhs();
                                }
                                else{
                                        return get_ubound(arraysym,i);
                                }
			break;
		case 2:
                                if(sub->rhs())
                                        return sub->rhs();
                                else  return new SgValueExp(1);
	}
}

SgExpression *indepdo_index_globalfunc(SgStatement * s,SgSymbol * index){
  Indepdo_header_def * ihdef = (Indepdo_header_def *) s->getAttribute(0, IHEDR_Def)->getAttributeData();

  for (int i=0; i<ihdef->getindexnumber(); i++) {
    if(ihdef->getindexsymb(i)->id()==index->id()) {

      Indepdo_index_def * iidef=(Indepdo_index_def *)s->attributeValue(i,IINDEX_Def);
      if(!iidef->loop_cut) continue;

      SgSymbol * rng=s->getSymbolInCurrentScope(mixstring, "irg", NULL, i);

      // we need to get local index;
      SgExpression * indexexp= new SgVarRefExp(index);

      SgSymbol * indexs = s->getSymbolInCurrentScope(mixstring, "lis", index->identifier());
      SgExpression * indexsexp= new SgVarRefExp(indexs);

      indexexp = &((* indexsexp) * (* indexexp));

      SgSymbol * indexl = s->getSymbolInCurrentScope(mixstring, "lil", index->identifier());
      SgExpression * indexlexp= new SgVarRefExp(indexl);

      indexexp = &((* indexexp) + (* indexlexp));

      SgExpression *globFunc = make_glob(s, rng, indexexp);

	return globFunc;	

      break;
    }
  }
}

int global_sub(SgStatement *s,SgCallStmt *call,Indepdo_header_def *ihdef,SgExpression *sub){
	int counter=0;
	for(int i=0;i<ihdef->getindexnumber();i++){
		SgSymbol *index=ihdef->getindexsymb(i);
		SgVarRefExp *indexExp=new SgVarRefExp(index);
		SgSymbol *global=s->getSymbolInCurrentScope(mixstring,"global",index->identifier());
		SgVarRefExp *globalExp=new SgVarRefExp(global);
		if(find_and_replace(sub,indexExp,globalExp)){
			counter++;
			SgStatement *stmt=new SgAssignStmt(*globalExp,*indepdo_index_globalfunc(s,index));
			call->insertStmtBefore(*stmt,*call->controlParent());
		}		
	}
	ASSERT(counter<2);
	return counter;
}

SgExpression *dim_len(SgSymbol * arraysym,int dim){
	SgExpression *len=&(* get_ubound(arraysym,dim) - * get_lbound(arraysym,dim) + * new SgValueExp(1));
	len=calculateExp(len);
	return len;
}

int get_template_dim(SgSymbol * arraysym,int dim){
	if(arraysym->numberOfAttributes(Align_Def)!=0){
		Alignment_def * adef = (Alignment_def *)(arraysym->attributeValue(0,Align_Def));
		return adef->getAlignPair(dim);
	}
	return dim;
}

SgExpression *dim_local_len(SgSymbol *tsym,Distribute_def *dist,int tdim){
	if(dist->distCode(tdim)==NOP) return dim_len(tsym,tdim);
	ASSERT(dist->distCode(tdim)==BLOCK||dist->distCode(tdim)==CYCLIC)
	int proc_dim=tdim;
	for(int i=0;i<tdim;i++){
		if(dist->distCode(i)==NOP) proc_dim--;
	}
	SgExpression * proc_dim_len=dist->getProcSize(proc_dim);
	SgExpression *exp=&(* dim_len(tsym,tdim) / * proc_dim_len);
	exp=calculateExp(exp);
	return exp;

}

SgExpression *dim_template_size(SgSymbol * arraysym,int dim){
	Distribute_def * dist=get_distdef(arraysym);
	if(!dist) return dim_len(arraysym,dim);
	SgSymbol *tsym=dist->getfromsymbol();
	
	return dim_local_len(tsym,dist,get_template_dim(arraysym,dim));
}

int get_nextdim(SgArrayRefExp *,int); 

SgExpression *dims_template_size(SgArrayRefExp *array,int dim1,int dim2){
	SgSymbol * arraysym=array->symbol();
	int maxdim=array->numberOfSubscripts();
	ASSERT(dim1>=0 && dim1<maxdim);
	ASSERT(dim2>=0 && dim2<maxdim);
	if(storageMode==COLUMN_MAJOR){ ASSERT(dim1<=dim2);}
	else { ASSERT(dim1>=dim2); }
	SgExpression *len=dim_template_size(arraysym,dim1);
	for(int dim=dim1;dim!=dim2;storageMode==COLUMN_MAJOR?dim++:dim--){
		len=&(*len * * dim_template_size(arraysym,storageMode==COLUMN_MAJOR?dim+1:dim-1));
	}
	return calculateExp(len);

}

SgExpression *template_size(SgArrayRefExp *array,int dim){
	int dim0;
	if(storageMode==COLUMN_MAJOR) dim0=0;
	else dim0=array->numberOfSubscripts()-1;
	return dims_template_size(array,dim0,dim);
}

SgExpression *calc_template(SgArrayRefExp *array,int dim,int dimcounter){
	int dim0;
	if(storageMode==COLUMN_MAJOR) dim0=0;
	else dim0=array->numberOfSubscripts()-1;
	if(dimcounter==0) return dims_template_size(array,dim0,get_nextdim(array,dim));
	else return dims_template_size(array,dim,get_nextdim(array,dim));
	
}

int get_lastdim(SgArrayRefExp *array,int dim);

SgExpression *template_stride(SgArrayRefExp *array,int dim,int dimcounter){
	int dim0;
	if(storageMode==COLUMN_MAJOR) dim0=0;
	else dim0=array->numberOfSubscripts()-1;
	SgExpression *t_str=new SgValueExp(1);
	if(dimcounter==0){
		int lastdim=get_lastdim(array,dim);
		if(dim!=lastdim)
			t_str=dims_template_size(array,dim0,lastdim);
	}	
	SgSymbol *arraysym=array->symbol();
	if(arraysym->numberOfAttributes(Align_Def)!=0){
                Alignment_def * adef = (Alignment_def *)(arraysym->attributeValue(0,Align_Def));
                t_str = &(*t_str * * adef->getStride(dim));
        }
	
	return calculateExp(t_str);
	
}

SgExpression *calc_template_multiplier(SgStatement *s,SgSymbol *arraysym,int refcounter,int last_dim,int dim){
	int tmp=storageMode==COLUMN_MAJOR?last_dim+1:last_dim-1;
	if(dim==tmp) return new SgValueExp(1);
//	SgSymbol *symb1=s->getSymbolInCurrentScope(mixstring,"u",NULL,refcounter,tmp);
//	SgSymbol *symb2=s->getSymbolInCurrentScope(mixstring,"u",NULL,refcounter,dim);
	SgSymbol *symb1=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),tmp+1);
	SgSymbol *symb2=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),dim+1);
	return &(*new SgVarRefExp(symb2) / *new SgVarRefExp(symb1));
}

SgExpression* make_indepdo_arg_range(SgStatement *s,SgCallStmt *call,int arg_num,int dim,SgExpression *start,SgExpression *end,SgSymbol *grp){
	SgFunctionSymb * prochedr;
        SgFunctionCallExp *proc_call;
        proc_call = (SgFunctionCallExp *)call->expr(0);
        prochedr=(SgFunctionSymb *)proc_call->rhs()->symbol();
	SgSymbol *dummy_sym=prochedr->parameter(arg_num);
	Alignment_def  *adef = NULL;
	Distribute_def *ddef = NULL;
	int proc_dim=0;
	if (dummy_sym->numberOfAttributes(Align_Def)!=0) {
	        adef = (Alignment_def *)(dummy_sym->attributeValue(0,Align_Def));
	        SgSymbol * t = adef->gettosymbol();
	        ddef = (Distribute_def *)t->attributeValue(0, Dist_Def);
		for(int i=0;i<dim;i++){
			if(ddef->distCode(i)==NOP);
			else proc_dim++;
		}
		switch(ddef->distCode(adef->getAlignPair(dim-1))){
			case NOP:
				return make_new_range_collapse(s,start,end);
			case BLOCK:
				return make_new_range_distribute(s,start,end,1,grp,proc_dim);
			case CYCLIC:
				return make_new_range_distribute(s,start,end,2,grp,proc_dim);
			default: printf("not surported distribution\n");
				ASSERT(0);
				exit(1);
		}
				
	}
	else {
	        if (dummy_sym->numberOfAttributes(Dist_Def) == 0) {
        		// INHERIT
		        if (dummy_sym->numberOfAttributes(Inherit_Def)!=0)
	                ;
        		else // default mapping (collapsed)
				return make_new_range_collapse(s,start,end);
        	}
        	else {
        		ddef = (Distribute_def *) dummy_sym->attributeValue(0, Dist_Def);
			for(int i=0;i<dim;i++){
				if(ddef->distCode(i)==NOP);
				else proc_dim++;
			}
			switch(ddef->distCode(dim-1)) {
				case NOP:
					return make_new_range_collapse(s,start,end);
				case BLOCK:
					return make_new_range_distribute(s,start,end,1,grp,proc_dim);
				case CYCLIC:
					return make_new_range_distribute(s,start,end,2,grp,proc_dim);
			}
        	}
  	}
}

int get_lastdim(SgArrayRefExp *array,int dim){
	int max_dim=array->numberOfSubscripts()-1;
	if(storageMode==COLUMN_MAJOR){
		if(dim==0) return dim;
		while(array->subscript(dim-1)->variant()!=DDOT){
			dim--;
			if(dim==0) return dim;
		}
		return dim;
	}
	else{
		if(dim==max_dim) return dim;
		while(array->subscript(dim+1)->variant()!=DDOT){
			dim++;
			if(dim==max_dim) return dim;
		}
		return dim;
	}
}
	
int get_nextdim(SgArrayRefExp *array,int dim){
	int max_dim=array->numberOfSubscripts()-1;
	if(storageMode!=COLUMN_MAJOR){
		if(dim==0) return dim;
		while(array->subscript(dim-1)->variant()!=DDOT){
			dim--;
			if(dim==0) return dim;
		}
		return dim;
	}
	else{
		if(dim==max_dim) return dim;
		while(array->subscript(dim+1)->variant()!=DDOT){
			dim++;
			if(dim==max_dim) return dim;
		}
		return dim;
	}
}

void indepdo_arg_array_create(SgStatement *s,SgCallStmt *call,int arg_num,int refcounter,SgSymbol *arg_dad,SgArrayRefExp *array){
	Indepdo_header_def *ihdef=(Indepdo_header_def *)s->attributeValue(0,IHEDR_Def);
	SgSymbol *arraysym=array->symbol();
	SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad",arraysym->identifier());
	SgFunctionCallExp *proc_call = (SgFunctionCallExp *)call->expr(0);
	SgFunctionSymb *prochedr=(SgFunctionSymb *)proc_call->rhs()->symbol();
	SgSymbol *dummysym=prochedr->parameter(arg_num);

	int arg_dad_dim=0;
	int dim=array->numberOfSubscripts();
	for(int i=0;i<dim;i++){
		if(array->subscript(i)->variant()==DDOT){
			arg_dad_dim++; 
		}
	}

	SgStatement *s1=s->controlParent();
	int variant=s1->variant();
	while(variant!=PROG_HEDR&&variant!=PROC_HEDR&&variant!=FUNC_HEDR&&variant!=MODULE_STMT){
		s1=s1->controlParent();
		variant=s1->variant();
	}
	ASSERT(s1);
	if(!s1->numberOfAttributes(Proc_Def)) ASSERT(0);
	SgSymbol *procs=((Processor_def *)s1->attributeValue(0,Proc_Def))->getsymbol();
	SgSymbol *grp=s->getSymbolInCurrentScope(mixstring,"grp",procs->identifier());
	SgStatement *new_array_stmt=make_new_array_data(s,arg_dad,arraysym->type(),-1,
arg_dad_dim,new SgVarRefExp(grp));

	s->addAttribute(INDEPDO_Coef,new_array_stmt,sizeof(new_array_stmt));

	int dimcounter=0;
	int counter=0;
	int last_dim;
	if(storageMode==COLUMN_MAJOR) last_dim=-1;
	else last_dim=dim;
	for(int loop=0;loop<dim;loop++){
		int i;
		if(storageMode==COLUMN_MAJOR) i=loop;
		else i=dim-1-loop;
		if(array->subscript(i)->variant()==DDOT){
			counter++;
			if(storageMode==COLUMN_MAJOR) dimcounter=counter;
			else dimcounter=arg_dad_dim-counter+1;

//make template of this dim
			SgExpression *start=new SgValueExp(1);//start of subscript and template of this dim
			SgSymbol *rng=s->getSymbolInCurrentScope(mixstring,"indepdo_rng","TEMP",refcounter,dimcounter);
		        SgExpression *rngExpr=new SgVarRefExp(rng);
			SgExpression *t_end=calc_template(array,i,counter-1);
			SgExpression *rngFunc=make_indepdo_arg_range(s,call,arg_num,dimcounter,start,t_end,grp);
			SgStatement *stmt=new SgAssignStmt(*rngExpr,*rngFunc);
			s->addAttribute(INDEPDO_Coef,stmt,sizeof(stmt));

//make align statement
			SgExpression *stride = ddot_sub_triplet(array,i,2);

			SgExpression *s_start=ddot_sub_triplet(array,i,0);
			SgExpression *end  = ddot_sub_triplet(array,i,1);
			SgExpression *s_end=&((*end-*s_start) / *stride+*start);
			analyse_expression(s,s_end);
			SgExpression *t_stride=&(*stride * * template_stride(array,i,counter-1));
			analyse_expression(s,t_stride);
			SgExpression *t_offset=&(* start - * start * *t_stride);
			analyse_expression(s,t_offset);
			stmt = make_set_array_align(s, arg_dad, dimcounter,start,s_end,t_stride,t_offset,new SgValueExp(0),new SgValueExp(0),rngExpr);
			s->addAttribute(INDEPDO_Coef,stmt,sizeof(stmt));
			stmt=make_delete_range(s,rngExpr);
			s->addAttribute(INSERTED_AFTER_STMT_IN,stmt,sizeof(stmt));
			last_dim=i;
		}
	}
}

void get_array_stride(SgStatement *s,SgArrayRefExp *array){
	SgSymbol *arraysym=array->symbol();
	if(arraysym->numberOfAttributes(STRIDE_GOT)) return;
	SgSymbol *dadsym=s->getSymbolInCurrentScope(mixstring,"dad",arraysym->identifier());
	SgExpression *dad=new SgVarRefExp(dadsym);
	SgArrayType *a_type=(SgArrayType *)arraysym->type();
	for(int j=0;j<a_type->dimension();j++){
		SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),j+1);
		SgExpression *stride=new SgVarRefExp(stridesym);
		SgSymbol *func=s->getSymbolInCurrentScope(mixstring,"mem_stride",NULL);
		SgFunctionCallExp *funcExp=new SgFunctionCallExp(*func);
		funcExp->addArg(*dad);
		funcExp->addArg(* new SgValueExp(j+1));
		SgAssignStmt *stride_init=new SgAssignStmt(*stride,*funcExp);
		s->insertStmtBefore(*stride_init,*s->controlParent());
	}
	arraysym->addAttribute(STRIDE_GOT,NULL,0);	
}

void dispose_indepdo_array_argument(SgStatement *s,int counter){
	Indepdo_reference *iref=(Indepdo_reference *)s->attributeValue(counter,Indepdo_Ref);
	SgArrayRefExp *array=(SgArrayRefExp *)iref->getexpression()->attributeValue(0,KEPT_ARRAY_REF);
	get_array_stride(s,array);
	SgSymbol *arraysym=array->symbol();
	SgSymbol *dad=s->getSymbolInCurrentScope(mixstring,"dad","arg",counter);
	SgVarRefExp *dadExp=new SgVarRefExp(dad);
	iref->getexpression()->addAttribute(Array_Arg_dad,dadExp,sizeof(dadExp));
	SgCallStmt *call=(SgCallStmt *)iref->getexpression()->attributeValue(0,Call_Stmt_of_arrayref);
	int arg_num=*(int *)iref->getexpression()->attributeValue(0,Array_Arg_num);
	indepdo_arg_array_create(s,call,arg_num,counter,dad,array);
	SgSymbol *zero=s->getSymbolInCurrentScope("0");
	SgStatement *stmt=make_reset_array_base1(s,dad,iref->getexpression(),zero);
	if(call->controlParent()->variant()==LOGIF_STMT) call->controlParent()->setVariant(LOGIF_NODE);
	call->insertStmtBefore(*stmt,*call->controlParent());
	stmt=make_delete_array(s,dad);
	s->addAttribute(INSERTED_AFTER_STMT_IN,stmt,sizeof(stmt));
}

void dispose_indepdo_array_argument1(SgStatement *s,int counter){
	Indepdo_reference *iref=(Indepdo_reference *)s->attributeValue(counter,Indepdo_Ref);
	SgArrayRefExp *array=(SgArrayRefExp *)iref->getexpression()->attributeValue(0,KEPT_ARRAY_REF);
	SgCallStmt *call=(SgCallStmt *)iref->getexpression()->attributeValue(0,Call_Stmt_of_arrayref);
	get_array_stride(s,array);
	SgSymbol *arraysym=array->symbol();
	SgSymbol *arg_stride=s->getSymbolInCurrentScope(mixstring,"arg","stride",counter);
	SgArrayRefExp *argstrideExp=new SgArrayRefExp(*arg_stride);
	iref->getexpression()->addAttribute(Array_Arg_dad,argstrideExp,sizeof(argstrideExp));
	int dim,arg_dim;
	arg_dim=-1;
	for(dim=0;dim<array->numberOfSubscripts();dim++){
		if(array->subscript(dim)->variant()==DDOT){
			arg_dim++;
			SgArrayRefExp *argstride=new SgArrayRefExp(*arg_stride);
			SgExpression *sub=new SgValueExp(arg_dim+1);
			sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
			argstride->addSubscript(*sub);
			SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),dim+1);
			SgExpression *mem_stride=new SgVarRefExp(stridesym);
			
			SgStatement *stmt=new SgAssignStmt(*argstride,* calculateExp(&(*mem_stride * *ddot_sub_triplet(array,dim,2))));
			s->addAttribute(INDEPDO_Coef,stmt,sizeof(stmt));
		}
	}
}

void trans_exp(SgStatement * s,SgExpression *expr){
  if(expr) {

    switch(expr->variant()) {
    case BOOL_VAL:
    case INT_VAL:
    case FLOAT_VAL:
    case DOUBLE_VAL:
    case STRING_VAL:
      break;

    case VAR_REF: {
      break;
    };

    case ARRAY_REF: {
	SgArrayRefExp *arrayExp=(SgArrayRefExp *)expr;
	SgSymbol *array=expr->symbol();
	if(is_dummy(s,array)){
		SgValueExp *zero=new SgValueExp(0);
		SgExpression *sub=zero;
		for(int i=0;i<arrayExp->numberOfSubscripts();i++){
			SgExpression *sub1;
			if(arrayExp->subscript(i)->variant()==DDOT) sub1=ddot_sub_triplet(arrayExp,i,0);
			else sub1=arrayExp->subscript(i);
			
			SgExpression *sub_start=get_lbound(array,i);
			SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"dim_stride",array->identifier(),i+1);
			SgExpression *stride=new SgVarRefExp(stridesym);
			sub=&(*sub + (*sub1 - *sub_start) * *stride);
		}
		sub=calculateExp(sub);
		sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
		SgArrayRefExp *newarray=new SgArrayRefExp(*array);
		newarray->addSubscript(*sub);
		arrayExp->replace(*newarray);
	}
      break;
    };

    case MINUS_OP:
    case ADD_OP:
    case MULT_OP:
    case DIV_OP:
    case EXP_OP:
    case MOD_OP:
    case AND_OP:
    case OR_OP:
    case EQV_OP:
    case NEQV_OP:

    case EQ_OP:
    case NOTEQL_OP:
    case LT_OP:
    case LTEQL_OP:
    case GT_OP:
    case GTEQL_OP: {
      trans_exp(s,expr->operand(1));
      trans_exp(s,expr->operand(2));
      break;
    }

    case NOT_OP: {
      trans_exp(s,expr->operand(1));
      break;
    }
	
    case FUNC_CALL:
    case INTRIN_FUNC_CALL: {
	break;
	}
    default:
      printf("Not processed expression in trans_exp!\n");
      ASSERT(0);
    };
  };
	
}

void get_proc_array_stride(SgStatement *s,SgArrayRefExp *array){
	SgSymbol *arraysym=array->symbol();
	if(is_dummy(s,arraysym)) return;

	get_array_stride(s,array);
}

void dispose_proc_argument(SgCallStmt *s,int argnum,int counter){
	SgArrayRefExp *array=(SgArrayRefExp *)s->arg(argnum);
	get_proc_array_stride(s,array);
	SgSymbol *arraysym=array->symbol();
	SgSymbol *arg_stride=s->getSymbolInCurrentScope(mixstring,"arg","stride",counter,argnum);
	SgArrayRefExp *argstrideExp=new SgArrayRefExp(*arg_stride);

	array->addAttribute(Array_Arg_dad,argstrideExp,sizeof(argstrideExp));

	SgStatement *s1=s->controlParent();
	SgStatement *s2=s;
        int variant=s1->variant();
        while(variant!=PROC_HEDR){
		s2=s1;
                s1=s1->controlParent();
                variant=s1->variant();
        }

	SgCallStmt *call=s;
	int dim,arg_dim;
	arg_dim=-1;
	SgArrayType *at=(SgArrayType *)arraysym->type();
	if(array->numberOfSubscripts())
	for(dim=0;dim<array->numberOfSubscripts();dim++){
		if(array->subscript(dim)->variant()==DDOT){
			arg_dim++;
			SgArrayRefExp *argstride=new SgArrayRefExp(*arg_stride);
			SgExpression *sub=new SgValueExp(arg_dim+1);
			sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
			argstride->addSubscript(*sub);
			SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),dim+1);
			SgExpression *mem_stride=new SgVarRefExp(stridesym);
			
			SgStatement *stmt=new SgAssignStmt(*argstride,* calculateExp(&(*mem_stride * *ddot_sub_triplet(array,dim,2))));
			s2->insertStmtBefore(*stmt,*s2->controlParent());
		}
	}
	else
	for(dim=0;dim<at->dimension();dim++){
		arg_dim++;
		SgArrayRefExp *argstride=new SgArrayRefExp(*arg_stride);
		SgExpression *sub=new SgValueExp(arg_dim+1);
		sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
		argstride->addSubscript(*sub);
		SgSymbol *stridesym=s->getSymbolInCurrentScope(mixstring,"dim_stride",arraysym->identifier(),dim+1);
		SgExpression *mem_stride=new SgVarRefExp(stridesym);
		
		SgStatement *stmt=new SgAssignStmt(*argstride,*mem_stride );
		s2->insertStmtBefore(*stmt,*s2->controlParent());
	}
	trans_exp(s,array);	
}

void trans_proc_call(SgCallStmt *call){
		SgSymbol *newname=call->getSymbolInCurrentScope(call->expr(0)->symbol()->identifier(),"indepdo",NULL);

                SgCallStmt *newcall=new SgCallStmt(*newname);
                int j;
                for(j=0;j<call->numberOfArgs();j++){
                        if(call->arg(j)->numberOfAttributes(Array_Arg_dad)==0){
                        newcall->addArg(*call->arg(j));
                        continue;
                        }
                        newcall->addArg(*call->arg(j));
                        SgVarRefExp *arg_stride=(SgVarRefExp *)call->arg(j)->attributeValue(0,Array_Arg_dad);
                        newcall->addArg(* arg_stride);
                }
                call->insertStmtBefore(*newcall,*call->controlParent());
                call->deleteStmt();
}

void indep_proc_trans(SgStatement *s,int counter){
	SgCallStmt * call=(SgCallStmt *)s;
        SgFunctionCallExp *proc_call = (SgFunctionCallExp *)call->expr(0);
	SgFunctionSymb *prochedr=(SgFunctionSymb *)proc_call->rhs()->symbol();
	for(int j=0;j<call->numberOfArgs();j++){
		if(prochedr->parameter(j)->type()->variant()==T_ARRAY){
			dispose_proc_argument(call,j,counter);		
		}

	}
	trans_proc_call(call);

}

SgStatement * forall2doloop(SgStatement *s){
	if(s->variant()!=FORALL_STMT) return s;
	SgForallStmt *forall=(SgForallStmt *)s;
	SgStatement *body=forall->childList1(0)->copyPtr();
	SgExpression *condition=forall->getMask();
	if(condition){
		body=new SgIfStmt(*condition,*body);
	}
	for(int level=0;level<forall->indexNumber();level++){
		SgSymbol *loopvar=forall->getIndexName(level);
		SgExpression *ddot=forall->getIndex(level)->rhs(); 
		body=new SgForStmt(*loopvar,*ddot->lhs()->lhs(),*ddot->lhs()->rhs(),*ddot->rhs(),*body);
	}
	s->insertStmtBefore(*body,*s->controlParent());
	s->deleteStmt();
	return body;
}

void indep_call_trans(SgStatement *s,int &proc_counter){
//  process_label(s);

  switch(s->variant()) {
  case GLOBAL:
    {
      break;
    };

  case PROG_HEDR:
    {
      break;
    }

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

	Flag kept=in_procedure;

      in_procedure=on;

      for (i=0; i<org_numberOfChildren; i++)
	indep_call_trans(children[i],proc_counter);

      delete [] children;
      
      in_procedure = off;

	in_procedure=kept;

	i=0;
	SgStatement *first_exec=s->childList1(i);
	while(first_exec->variant()==VAR_DECL||first_exec->variant()==PARAM_DECL){
	i++;
	first_exec=s->childList1(i);
	}
	 for(i=0;i<prochedr->numberOfParameters();i++){
		if(prochedr->parameter(i)->type()->variant()==T_ARRAY){
			if(prochedr->parameter(i)->numberOfAttributes(STRIDE_GOT)) continue;
			SgArrayType *a_type=(SgArrayType *)prochedr->parameter(i)->type();
			for(int j=0;j<a_type->dimension();j++){
			SgSymbol *stridesym=first_exec->getSymbolInCurrentScope(mixstring,"dim_stride",prochedr->parameter(i)->identifier(),j+1);
			SgExpression *stride=new SgVarRefExp(stridesym);
    			SgSymbol * arg_stride = s->getSymbolInCurrentScope(mixstring, "arg_stride", prochedr->parameter(i)->identifier());
			SgArrayRefExp *argstride=new SgArrayRefExp(*arg_stride);
			SgExpression *sub=new SgValueExp(j+1);
			sub=new SgExpression(EXPR_LIST,sub,NULL,NULL);
			argstride->addSubscript(*sub);
			SgAssignStmt *stride_init=new SgAssignStmt(*stride,*argstride);
			first_exec->insertStmtBefore(*stride_init,*s);
			}
			prochedr->parameter(i)->addAttribute(STRIDE_GOT,NULL,0);	
		}
	 }
      break;
    };

  case FUNC_HEDR:
    {
      break;
    };

  case MODULE_STMT:
    {
      in_module = on;

      in_module=off;

      break;
    }

  case CONTAINS_STMT:
    break;

  case INTERFACE_STMT:
    s->deleteStmt();
    break;

  case PROCESSORS_STMT:
    s->deleteStmt();
    break;

  case TEMPLATE_STMT:
    s->deleteStmt();
    break;

  case DYNAMIC_DECL:
    s->deleteStmt();
    break;

  case ALIGN_DECL:
    s->deleteStmt();
    break;

  case REALIGN_STMT:
    s->deleteStmt();
    break;

  case DISTRIBUTE_DECL:
    s->deleteStmt();
    break;

  case REDISTRIBUTE_STMT:
    s->deleteStmt();
    break;

  case INHERIT_DECL:
    s->deleteStmt();
    break;

  case OUTOFCORE_DECL:
  case INCORESIZE_DECL:
  case MEMORY_DECL:
  case MAJOR_DECL:
    s->deleteStmt();
    break;

  case FILEMAP_STMT:
    s->deleteStmt();
    break;

  case IO_PROCESSOR_STMT:
    s->deleteStmt();
    break;

  case SHADOW_STMT:
    s->deleteStmt();
    break;

  case VAR_DECL:
    delete_dummy_array(s);
    break;

  case INDEPENDENT_STMT:
    s->deleteStmt();
    break;

  case FOR_NODE:
    {
      do_level++;

      SgExpression * doHeaderExp=s->expr(0);
//      analyse_expression(s, doHeaderExp);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
        indep_call_trans(children[i],proc_counter);
      
      delete [] children;

      do_level--;
      
      break;
    }

  case DO_WHILE_NODE:
    {
      do_level++;

      SgExpression * doHeaderExp=s->expr(0);
//      analyse_expression(s, doHeaderExp);

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
        indep_call_trans(children[i],proc_counter);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
    {
      if_level ++;

      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	indep_call_trans(children1[i],proc_counter);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);

      for (i=0; i<org_numberOfChildren2; i++)
	indep_call_trans(children2[i],proc_counter);
      
      delete [] children2;

	// LOGIF_STMT => LOGIF_NODE
      if (s->variant() == LOGIF_STMT)
      {
	s->setVariant(LOGIF_NODE);
      }

      if_level --;

      break;
    }

  case SWITCH_NODE: {
	break;
    }
  case CASE_NODE:
    {
      break;
    }

  case SIMPLE_ON_STMT:
	s->deleteStmt();
      break;

  case BLOCK_ON_STMT: {
	s->deleteStmt();
      break;

    }
    break;

  case SIMPLE_RESIDENT_STMT:
	s->deleteStmt();
      break;
      break;

  case BLOCK_RESIDENT_STMT: {
	s->deleteStmt();
      break;

    }
    break;

  case TASK_STMT: {

    }
    break;

  case ASSGOTO_NODE:
    break;

  case STOP_STAT:
    break;

  case PROC_STAT:
	indep_proc_trans(s,proc_counter);
	proc_counter++;
    break;

  case ASSIGN_STAT:
	trans_exp(s,s->expr(0));
	trans_exp(s,s->expr(1));
    break;

  case POINTER_ASSIGN_STAT:
    break;

  case FORALL_STMT:
	forall2doloop(s);
    break;

  case OPEN_STAT:
  case CLOSE_STAT:
    break;

  case READ_STAT:
  case PRINT_STAT:
  case WRITE_STAT:
    break;

  case DIR_END:
    break;

  case CONTROL_END:
  case CONTINUE_NODE:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
    break;

  case RETURN_STAT:
   break;

  case INTRIN_STAT:
    s->deleteStmt();
   break;

  case EXIT_STMT:
  case IMPL_DECL:
  case EXTERN_STAT:
  case COMM_STAT:
   // statement need do nothing;
   break;

  case ALLOCATABLE_STMT:
    s->deleteStmt();
    break;

  case ALLOCATE_STMT:
    break;

  case DEALLOCATE_STMT:
    break;

  default:
    break;
  };
}
